from collections.abc import AsyncIterator
from typing import Any

from nextline.events import OnWriteStdout
from nextline.plugin.spec import hookimpl
from nextline.utils.pubsub import PubSubItem


class CacheStdout:
    def __init__(self) -> None:
        self._pubsub = PubSubItem[str](cache=True)

    async def subscribe(self) -> AsyncIterator[str]:
        async for text in self._pubsub.subscribe():
            yield text

    async def aclose(self) -> None:
        await self._pubsub.aclose()

    async def __aenter__(self) -> 'CacheStdout':
        return self

    async def __aexit__(self, *_: Any, **__: Any) -> None:
        await self.aclose()

    @hookimpl
    async def on_initialize_run(self) -> None:
        self._pubsub.clear()

    @hookimpl
    async def on_write_stdout(self, event: OnWriteStdout) -> None:
        await self._pubsub.publish(event.text)



'''Strawberry GraphQL ASGI app with a fix for async_asgi_testclient.

Strawberry: https://strawberry.rocks
'''

from typing import TYPE_CHECKING

from strawberry.asgi import GraphQL as GraphQL_
from strawberry.subscriptions import GRAPHQL_WS_PROTOCOL

if TYPE_CHECKING:
    from strawberry.asgi import Receive, Scope, Send


class GraphQL(GraphQL_):
    """Add a fix to the strawberry GraphQL for async_asgi_testclient"""

    async def __call__(self, scope: 'Scope', receive: 'Receive', send: 'Send') -> None:
        if scope["type"] == "websocket":
            if not scope.get("subprotocols"):  # pragma: no branch
                # strawberry closes the websocket connection if
                # subprotocols are empty, which is the case for
                # async_asgi_testclient.
                scope["subprotocols"] = [GRAPHQL_WS_PROTOCOL]
        return await super().__call__(scope, receive, send)



import sys
from pathlib import Path

__all__ = ["statement"]

THIS_DIR = Path(__file__).resolve().parent
SCRIPT_PATH = THIS_DIR / 'script.py'

sys.path.insert(0, str(THIS_DIR))

with open(SCRIPT_PATH) as f:
    statement = f.read()



from pathlib import Path

from graphql import parse, print_ast


def read_gql(path: Path | str) -> str:
    '''Load a GraphQL query from a file while checking its syntax.'''

    text = Path(path).read_text()
    parsed = parse(text)
    reformatted = print_ast(parsed)
    return reformatted


pwd = Path(__file__).resolve().parent

sub = pwd / 'mutations'
MUTATE_EXEC = read_gql(sub / 'Exec.gql')
MUTATE_RUN_AND_CONTINUE = read_gql(sub / 'RunAndContinue.gql')
MUTATE_RESET = read_gql(sub / 'Reset.gql')
MUTATE_SEND_PDB_COMMAND = read_gql(sub / 'SendPdbCommand.gql')
MUTATE_INTERRUPT = read_gql(sub / 'Interrupt.gql')
MUTATE_TERMINATE = read_gql(sub / 'Terminate.gql')
MUTATE_KILL = read_gql(sub / 'Kill.gql')
MUTATE_LOAD_EXAMPLE_SCRIPT = read_gql(sub / 'LoadExampleScript.gql')

sub = pwd / 'queries'
QUERY_STATE = read_gql(sub / 'State.gql')
QUERY_RUN_NO = read_gql(sub / 'RunNo.gql')
QUERY_TRACE_IDS = read_gql(sub / 'TraceIds.gql')
QUERY_SOURCE = read_gql(sub / 'Source.gql')
QUERY_SOURCE_LINE = read_gql(sub / 'SourceLine.gql')
QUERY_EXCEPTION = read_gql(sub / 'Exception.gql')
QUERY_CONTINUOUS_ENABLED = read_gql(sub / 'ContinuousEnabled.gql')

sub = pwd / 'subscriptions'
SUBSCRIBE_COUNTER = read_gql(sub / 'Counter.gql')
SUBSCRIBE_STATE = read_gql(sub / 'State.gql')
SUBSCRIBE_RUN_NO = read_gql(sub / 'RunNo.gql')
SUBSCRIBE_TRACE_IDS = read_gql(sub / 'TraceIds.gql')
SUBSCRIBE_PROMPTING = read_gql(sub / 'Prompting.gql')
SUBSCRIBE_STDOUT = read_gql(sub / 'Stdout.gql')
SUBSCRIBE_CONTINUOUS_ENABLED = read_gql(sub / 'ContinuousEnabled.gql')



from collections.abc import AsyncIterator, MutableMapping
from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Any, Optional

import strawberry
from apluggy import PluginManager
from dynaconf import Dynaconf
from starlette.applications import Starlette
from starlette.types import ASGIApp
from strawberry.schema import BaseSchema
from strawberry.tools import merge_types

from nextlinegraphql.custom.strawberry import GraphQL
from nextlinegraphql.hook import spec

from .schema import Query

if TYPE_CHECKING:
    from strawberry.asgi import Request, Response, WebSocket


class Plugin:
    @spec.hookimpl
    def configure(self, settings: Dynaconf, hook: PluginManager) -> None:
        self._settings = settings
        self._app = _create_app(hook=hook)

    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, None, None)

    @spec.hookimpl(tryfirst=True)  # tryfirst so to be the outermost context
    @asynccontextmanager
    async def lifespan(self, app: Starlette) -> AsyncIterator[None]:
        app.mount('/', self._app)
        yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping[str, Any]) -> None:
        context['settings'] = self._settings


def _create_app(hook: PluginManager) -> ASGIApp:
    schema = _compose_schema(hook=hook)
    app = _EGraphQL(schema).set_hook(hook)
    return app


def _compose_schema(hook: PluginManager) -> BaseSchema:
    # [(Query, Mutation, Subscription), ...]
    three_types = hook.hook.schema()

    # [(Query, ...), (Mutation, ...), (Subscription, ...)]
    transposed = list(map(tuple, zip(*three_types)))
    transposed = [tuple(t for t in l if t) for l in transposed]  # remove None

    q, m, s = transposed
    assert q  # Query is required
    Query = merge_types('Query', q)  # type: ignore
    Mutation = m and merge_types('Mutation', m) or None  # type: ignore
    Subscription = s and merge_types('Subscription', s) or None  # type: ignore

    schema = strawberry.Schema(
        query=Query,
        mutation=Mutation,
        subscription=Subscription,
    )

    return schema


class _EGraphQL(GraphQL):
    '''Extend the strawberry GraphQL app to override the `get_context` method

    This class is implemented in the way described in the strawberry document:
    https://strawberry.rocks/docs/integrations/asgi
    '''

    def set_hook(self, hook: PluginManager) -> '_EGraphQL':
        self._hook = hook
        return self

    async def get_context(
        self, request: 'Request | WebSocket', response: 'Response | WebSocket'
    ) -> Optional[Any]:
        context = {'request': request, 'response': response}
        self._hook.hook.update_strawberry_context(context=context)
        return context



import contextlib
import logging.config
from collections.abc import AsyncIterator
from logging import getLogger
from typing import Any

from rich import print
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware

from .config import load_settings
from .hook import load_plugins


def create_app() -> Starlette:
    '''App factory for Uvicorn

    Use the factory option to run.

    $ uvicorn --factory nextlinegraphql:create_app

    Uvicorn Doc: https://www.uvicorn.org/#application-factories
    '''

    hook = load_plugins()
    config = load_settings(hook)
    print('Settings:', config.as_dict())

    configure_logging(config.logging)
    logger = getLogger(__name__)
    logger.info('Logging configured')

    hook.hook.configure(settings=config, hook=hook)

    plugin_names = [n for n, p in hook.list_name_plugin() if p]
    logger.info(f'Pluggy project name: {hook.project_name!r}')
    logger.info(f'Loaded plugins: {plugin_names}')

    @contextlib.asynccontextmanager
    async def lifespan(app: Starlette) -> AsyncIterator[None]:
        context = dict[Any, Any]()
        await hook.ahook.update_lifespan_context(app=app, hook=hook, context=context)
        async with hook.awith.lifespan(
            app=app, hook=hook, context=context
        ):  # pragma: no branch
            yield

    middleware = [
        Middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_methods=["*"],
            allow_headers=["*"],
        )
    ]

    app = Starlette(debug=True, lifespan=lifespan, middleware=middleware)

    return app


def configure_logging(config: dict) -> None:
    logging.config.dictConfig(config)

    # https://pypi.org/project/logging_tree/
    # import logging_tree
    # logging_tree.printout()



from pathlib import Path

from graphql import parse, print_ast


def read_gql(path: Path | str) -> str:
    '''Load a GraphQL query from a file while checking its syntax.'''

    text = Path(path).read_text()
    parsed = parse(text)
    reformatted = print_ast(parsed)
    return reformatted


pwd = Path(__file__).resolve().parent


sub = pwd / 'queries'
QUERY_HEADERS = read_gql(sub / 'Headers.gql')



__all__ = ['Plugin']

from collections.abc import AsyncIterator, MutableMapping
from contextlib import asynccontextmanager

from nextline import Nextline
from nextlinegraphql.hook import spec

from .cache import CacheStdout
from .example_script import statement
from .schema import Mutation, Query, Subscription


class Plugin:
    @spec.hookimpl
    def schema(self) -> tuple[type, type | None, type | None]:
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    async def update_lifespan_context(self, context: MutableMapping) -> None:
        self._nextline = Nextline(statement)
        context['nextline'] = self._nextline

    @spec.hookimpl(trylast=True)  # trylast so to be the innermost context
    @asynccontextmanager
    async def lifespan(self) -> AsyncIterator[None]:
        '''Yield within the nextline context.'''
        self._cache_stdout = CacheStdout()
        self._nextline.register(self._cache_stdout)
        async with self._cache_stdout, self._nextline:
            yield

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        context['nextline'] = self._nextline
        ctrl = {'cache_stdout': self._cache_stdout}
        context['ctrl'] = ctrl



"""
uvicorn --factory --reload --reload-dir nextline-graphql --reload-dir nextline nextlinegraphql:create_app
use "--lifespan on" to show exception in lifespan
"""

__all__ = ['__version__', 'create_app', 'spec']

import sys  # noqa: E402

from .__about__ import __version__
from .factory import create_app
from .hook import spec

# remove args so that they won't be processed in the executing script
# TODO This should be properly handled and tested
sys.argv[1:] = []



import time

time.sleep(0.3)

def f():
    for _ in range(10):
        pass
    return

f()
f()

print('here!')

import script_threading

script_threading.run()


import script_asyncio

script_asyncio.run()
